<?php

/**
 * UnusedImagesUsageFinderVariables looks for references to images in variables.
 *
 * Where to look for variables:
 * 1. Table variable contains serialized values.
 * 2. Or, maybe even better as it also contains settings from settings.php and
 *    contains already unserialized values, global $conf.
 * 3. Translation: support the variable_store module with its variable_store
 *    table: strings or serialized values.
 *
 * How to recognize references to images:
 * We cannot use the same restriction as in text fields, namely that an image
 * should appear as an attribute value and thus appears between (double) quotes.
 * We also cannot use the method the url filter uses, see _filter_url().
 *
 * So we have do it differently:
 * - Find all occurrences of one of the selected extensions preceded by a dot.
 * - Process all these occurrences by extending these matches:
 *   * backward: with all characters accepted
 *   * forward: with all characters accepted
 *
 * Note: https://perishablepress.com/stop-using-unsafe-characters-in-urls/ was
 * used to get the initial list of acceptable characters. But contrary to its
 * advice, it is common usage to allow unicode characters outside the ascii
 * range as browsers will encode them when sending them over the http protocol.
 * So I compiled a list of non-acceptable characters instead of acceptable
 * characters. Of course unsafe/reserved characters with a special meaning (e.g.
 * ':' and '/') are to be allowed as well. Whereas the safe character "'" is
 * often used as delimiter, so will not be accepted.
 */
class UnusedImagesUsageFinderVariables extends UnusedImagesUsageFinderTableColumn {

  /**
   * UnusedImagesUsageFinderVariables constructor.
   *
   * @param array $extensions
   */
  public function __construct(array $extensions) {
    $this->setUsageType('Variables');
    parent::__construct($extensions);
  }

  /**
   * {@inheritDoc}
   */
  protected function execute(array $settings) {
    $this->initSearch($settings);
    $results = $this->findUsages('variable', 'value', array('name'), $this->extensions);
    if (module_exists('variable_store')) {
      $results = (new UnusedImagesUtilities())->mergeUsageResults($results,
        $this->findUsages('variable_store', 'value', array('realm', 'realm_key', 'name'), $this->extensions));
    }
    return $results;
  }

  protected function initSearch(array $settings) {
    parent::initSearch($settings);

    // Characters not accepted:
    // ASCII Control characters	00-1F hex (0-31 decimal) and 7F (127 decimal.)
    $asciiControl = '';
    for ($i = 0; $i < 32; $i++) {
      $asciiControl .= chr($i);
    }
    $asciiControl .= chr(127);

    // Reserved characters: ; / ? : @ = &
    // Of these, '@', ':' and '/' have a special meaning in building a uri and
    // thus should be accepted. '?', '=' and '&' have a special meaning in the
    // query part, which we ignore, so we will not accept these.
    $reserved = ';?=&@';

    // Unsafe characters: space and " < > # % { } | \ ^ ~ [ ] `
    // We add ' as this will probably be a delimiter of the uri, not part of it.
    // We also add * as it is forbidden in file names on Windows and, if found
    // anyway, probably serves as search wildcard.
    $unsafe = ' "<>#%{}|\\^~[]`\'';

    // List of accepted characters is every character except those listed in
    // this regular expression notation.
    $accepted = '[^\\x00-\\x1F\\x7F;?=&@* "<>#%{}|\\\\^~[\\]`\']';
    $regExpExtensions = '\.(' . implode('|', $settings['extensions']) . ')';

    $this->regExps = array();
    $this->regExps[] = "/($accepted+$regExpExtensions)/";
  }

}
