<?php

/**
 * Class UnusedImagesTextFields
 */
class UnusedImagesUtilities {

  /**
   * Returns a list of fields that may contain textual references to image URIs.
   *
   * @return string[]
   *   Array of field labels that are keyed by the machine name of the field
   *   name. Field labels look like:
   *   "field_name (type) used in: entity_type[ (bundle[, bundle, ...])]"
   */
  public function getFieldsThatMayBeSearched() {
    $options = array();
    $field_types = $this->getFieldTypesThatMayBeSearched();
    foreach ($field_types as $field_type) {
      $options = array_merge($options, $this->getFieldsByType($field_type));
    }
    return $options;
  }

  /**
   * Returns field types that may contain textual references to image URIs.
   *
   * @return string[]
   *   List of of field types (machine name).
   */
  private function getFieldTypesThatMayBeSearched() {
    $options = array();
    if (module_exists('text')) {
      $options[] = 'text';
      $options[] = 'text_long';
      $options[] = 'text_with_summary';
    }
    if (module_exists('link')) {
      $options[] = 'link_field';
    }
    return $options;
  }

  /**
   * Returns all fields that are of the given type.
   *
   * @param string $field_type
   *
   * @return string[]
   *   Array of field labels that are keyed by the machine name of the field
   *   name.
   */
  private function getFieldsByType($field_type) {
    $result = array();
    $fields = field_info_fields();
    foreach ($fields as $field_info) {
      if ($field_info['type'] === $field_type) {
        $label = $field_info['field_name'] . ' (' . $field_info['type'] . ') ' . t('used in:');
        foreach ($field_info['bundles'] as $entity_type => $bundles) {
          $label .= ' ';
          if (count($bundles) === 1 && $entity_type === reset($bundles)) {
            $label .= $entity_type;
          }
          else {
            $label .= $entity_type . ' (' . implode(', ', $bundles) . ')';
          }
        }
        $result[$field_info['field_name']] = $label;
      }
    }
    return $result;
  }

  /**
   * Merges 2 sets of usage results.
   *
   * A simple array_merge() doesn't suffice:
   * -
   * - The set to merge into is keyed by image. The set to merge is not keyed by
   *   image.
   * - When merging results, the sets of primary keys that refer to the same
   *   image needs to be merged using specific logic (especially the isMultiple
   *   property and its consequences for what type of an array (set of keys or
   *   set of key-fields) the property keys will contain.
   *
   * @param array $results
   *   The 1st set of results, keyed by image
   * @param array $newUsages
   *   The 2nd set of results, not keyed.
   * @param \stdClass|null $result
   *   The result object for the 2nd set of results containing meta info that
   *   may be added to each separate result of that 2nd set.
   *
   * @return array
   *   The merged results.
   */
  public function mergeUsageResults(array $results, array $newUsages, stdClass $result = null) {
    foreach ($newUsages as $newUsage) {

      $uri = $newUsage->image;
      if (isset($results[$uri])) {
        if (!$results[$uri]->isMultiple) {
          $results[$uri]->keys = array($results[$uri]->keys);
          $results[$uri]->isMultiple = TRUE;
        }
        if (!$newUsage->isMultiple) {
          $newUsage->keys = array($newUsage->keys);
        }
        $results[$uri]->keys = array_merge($results[$uri]->keys, $newUsage->keys);
      }
      else {
        $results[$uri] = $newUsage;
      }
    }
    return $results;
  }

  /**
   * Converts a result to a human readable string.
   *
   * @param string|\stdClass $result
   *
   * @return string
   */
  public function resultToString($result) {
    if (is_object($result)) {
      $keys = $result->keys;
      if ($result->isMultiple) {
        $usageCount = count($keys);
        $keys = array_map(function($key) {
          if (is_array($key)) {
            $key = implode(',', $key);
          }
          $key = "[$key]";
          return $key;
        }, $keys);
      }
      else {
        $usageCount = 1;
        if (is_array($keys)) {
          $keys = implode(',', $keys);
        }
        $keys = "[$keys]";
      }
      if (is_array($keys)) {
        $keys = implode(',', $keys);
      }
      if (strlen($keys) > 96) {
        $keys = substr($keys, 0, 93) . '...';
      }
      return sprintf('<span class="image-found"">%s:</span> <span class="usages-found"">%d usages: %s</span>', $result->image, $usageCount, $keys);
    }
    else {
      return (string) $result;
    }
  }

  /**
   * Helper method that captures knowledge of specific field types.
   *
   * @param string $field_type
   *
   * @return string[]
   *   A list of field columns to search for occurrences of the file names.
   */
  public function getColumnsByFieldType($field_type) {
    switch ($field_type) {
      case 'link_field':
        $result = array('url');
        break;
      case 'text':
        $result = array('value');
        break;
      case 'text_long':
        $result = array('value');
        break;
      case 'text_with_summary':
        $result = array('value', 'summary');
        break;
      default:
        $result = array('value');
        break;
    }
    return $result;
  }

}
